<?php

use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\user_history\Entity\UserHistory;
use Drupal\Core\Url;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\File\FileSystemInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Function to add one or more fields to the user_history entity to allow
 * tracking of changes to the same field on the user entity.  This function
 * will be executed in a batch context.
 *
 * @param array $tracked_fields
 * @param array $context
 *
 * @throws \Drupal\Core\Entity\EntityStorageException
 */
function user_history_add_tracked_fields($tracked_fields, &$context) {

  // Get the field definitions for fields attached to user entity.
  $field_definitions = \Drupal::service('entity_field.manager')->getFieldDefinitions('user', 'user');

  // Get the field config from user fields for each field to be added,
  // and add fields to user_history entity.
  foreach($tracked_fields as $field_name) {

    /** @var \Drupal\field\Entity\FieldConfig $user_field */
    $user_field = $field_definitions[$field_name];

    $field_storage = FieldStorageConfig::loadByName('user_history', $field_name);
    if (empty($field_storage)) {
      // Get field storage values from user field and adjust for user_history.
      $field_storage_values = $user_field->getFieldStorageDefinition()->toArray();
      unset($field_storage_values['uuid']);
      unset($field_storage_values['id']);
      $field_storage_values['entity_type'] = 'user_history';
      $field_storage_values['module'] = 'user_history';
      $field_storage_values['dependencies'] = [
        'module' => ['user_history'],
      ];
      // Create the field storage config entity.
      $field_storage = FieldStorageConfig::create($field_storage_values);
      $field_storage->save();
    }

    $field_config = FieldConfig::loadByName('user_history', 'user_history', $field_name);
    if (empty($field_config)) {
      // Get field instance values from user field and adjust for user_history.
      $field_config_values = $user_field->toArray();
      unset($field_config_values['uuid']);
      unset($field_config_values['id']);

      $field_config_values['field_storage'] = $field_storage;
      $field_config_values['entity_type'] = 'user_history';
      $field_config_values['bundle'] = 'user_history';
      $field_config_values['dependencies'] = [
        'config' => ['field.storage.' . $field_storage->id()],
        'module' => ['user_history'],
      ];
      // Create the field instance config entity.
      $field_config = FieldConfig::create($field_config_values);
      $field_config->save();
    }

    // Log message about new field added.
    \Drupal::logger('user_history')->info('Field %label (%name) added to user_history entity.',
      ['%label' => $field_config->getLabel(), '%name' => $field_name]);

  }

  // Update the details of the running config
  $config = \Drupal::config('user_history.settings');
  \Drupal::state()->set('user_history.base_fields', $config->get('base_field'));
  \Drupal::state()->set('user_history.attached_fields', $config->get('attached_field'));

}

/**
 * Function to remove a field from the user_history entity which is no longer
 * required to be tracked.  This function will be executed in a batch context.
 *
 * @param array $remove_fields
 * @param array $context
 *
 * @throws \Drupal\Core\Entity\EntityStorageException
 */
function user_history_remove_tracked_fields($remove_fields, &$context) {
  \Drupal::messenger()->addMessage(t('Remove fields from user_history: %fields', ['%fields'=>implode(', ', $remove_fields)]));

  // Delete the field instance config and field storage config for each field to be removed.
  foreach($remove_fields as $field_name) {

    $field_config = FieldConfig::loadByName('user_history', 'user_history', $field_name);
    if (!empty($field_config)) {
      $field_config->delete();
    }

    $field_storage = FieldStorageConfig::loadByName('user_history', $field_name);
    if (!empty($field_storage)) {
      $field_storage->delete();
    }

    // Log message about existing field removed.
    \Drupal::logger('user_history')->info('Field %label (%name) removed from user_history entity.',
      ['%label' => $field_config->getLabel(), '%name' => $field_name]);

  }

  // Update the details of the running config
  $config = \Drupal::config('user_history.settings');
  \Drupal::state()->set('user_history.base_fields', $config->get('base_field'));
  \Drupal::state()->set('user_history.attached_fields', $config->get('attached_field'));

}

/**
 * Function to create initial user_history records for all users currently in
 * the database.  This function will be executed in a batch context.
 *
 * @param integer $batch_size
 * @param array $context
 *
 * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
 * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
 * @throws \Drupal\Core\Entity\EntityStorageException
 */
function user_history_create_initial_history($batch_size, &$context) {

  $entity_type = 'user_history';
  $values = [];
  $current_user = \Drupal::currentUser();

  $config = \Drupal::config('user_history.settings');
  // Get a list of fields that should be tracked by the user_history entity.
  $attached_fields = $config->get('attached_field');
  $tracked_fields = [];
  foreach ($attached_fields as $field_name => $tracked) {
    if ($tracked) {
      $tracked_fields[] = $field_name;
    }
  }

  // Initialise the sandbox values on the first call.
  if (empty($context['sandbox'])) {
    $context['sandbox'] = [];
    $context['sandbox']['total'] = count(\Drupal::entityQuery('user')
      ->condition('uid', 0, '>')
      ->execute());
    $context['sandbox']['last'] = 0;
    $context['sandbox']['count'] = 0;
    // Get current time for entitiy labels
    $now = new \Drupal\Core\Datetime\DrupalDateTime('now', drupal_get_user_timezone());
    $context['sandbox']['date'] = $now->format('Ymd-Hi');
    // Set the install date.
    \Drupal::state()->set('user_history.install_date', $now->format('Y-m-d H:i'));
  }

  // Retrieve the next set of user accounts
  $uids = \Drupal::entityQuery('user')
    ->condition('uid', $context['sandbox']['last'], '>')
    ->range(0, $batch_size)
    ->execute();

  $accounts = \Drupal::entityTypeManager()->getStorage('user')->loadMultiple($uids);

  /** @var \Drupal\user\Entity\User $account */
  foreach ($accounts as $account) {
    // Ignore the anonymous account.
    if (!$account->isAnonymous()) {

      $user_history = new UserHistory($values, $entity_type);

      // Set the label for this user_history entity
      $user_history->setLabel($account->id() . ': Install - ' . $context['sandbox']['date']);

      // Set the modified time to the last changed time of the account
      $user_history->setCreatedTime($account->getChangedTime());
      // Set the modified_by uid value to current user.
      $user_history->setModifiedBy($current_user);
      // Record that this record was generated during module install.
      $user_history->setAction('Install');
      // Record that this user entity exists in the database.
      $user_history->setUserDeleted(FALSE);

      $user_history->setUserId($account->id());
      $user_history->setUsername($account->getAccountname());
      $user_history->setUserPass($account->getPassword());
      $user_history->setUserMail($account->getEmail());
      $user_history->setUserTimezone($account->getTimeZone());
      $user_history->setUserStatus($account->isActive());
      $user_history->setUserRoles(implode('; ', $account->getRoles(TRUE)));
      $user_history->setUserCreated($account->getCreatedTime());
      $user_history->setUserChanged($account->getChangedTime());
      $user_history->setUserAccess($account->getLastAccessedTime());
      $user_history->setUserLogin($account->getLastLoginTime());
      $user_history->setUserInit($account->getInitialEmail());
      $user_history->setUserLangcode($account->language());
      $user_history->setUserPreferredLangcode($account->getPreferredLangcode());
      $user_history->setUserPreferredAdminLangcode($account->getPreferredAdminLangcode());

      // Get a list of fields attached to the user_history entity.
      $field_definitions = \Drupal::service('entity_field.manager')->getFieldDefinitions('user_history', 'user_history');
      // Add values of tracked fields attached to the user entity.
      foreach($tracked_fields as $field_name) {

        /** \Drupal\field\Entity\FieldConfig $field_definition */
        $field_definition = $field_definitions[$field_name];
        $field_type = $field_definition->getType();

        /** @var \Drupal\Core\Field\FieldItemList $field_item_list */
        $field_item_list = $account->get($field_name);

        $history_values = user_history_get_tracked_field_value($field_item_list);
        $user_history->set($field_name, $history_values);
      }
      // Mark this user_history as a new entity.
      $user_history->enforceIsNew(TRUE);

      $user_history->save();

    }

    $context['sandbox']['last'] = $account->id();
    $context['sandbox']['count']++;
  }

  if ($context['sandbox']['count'] < $context['sandbox']['total']) {
    $context['finished'] = $context['sandbox']['count'] / $context['sandbox']['total'];
    if ($context['finished'] > 0.99) {
      $context['finished'] = 0.99;
    }
    $context['message'] = t('Processed %count of %total user accounts.', [
      '%count' => $context['sandbox']['count'],
      '%total' => $context['sandbox']['total'],
    ]);
  }
  else {
    $context['finished'] = 1;
    $context['message'] = t('Finished processing %total user accounts.', [
      '%total' => $context['sandbox']['total'],
    ]);
    $context['results']['total'] = $context['sandbox']['total'];
  }

}

/**
 * Completion function for batch operation to initialise the user_history
 * records.
 *
 * @param boolean $success
 * @param array $results
 * @param array $operations
 *
 * @return \Symfony\Component\HttpFoundation\RedirectResponse
 */
function user_history_finished_initial_history($success, $results, $operations) {

  $messenger = \Drupal::messenger();

  if ($success) {
    // Flag the initialisation as complete.
    \Drupal::state()->set('user_history.initialise_required',FALSE);
    // Display a completion message to the user.
    \Drupal::messenger()->addMessage(t('Added %total new user_history records', ['%total' => $results['total']]));
  }
  else {
    // An error occurred.
    // $operations contains the operations that remained unprocessed.
    $error_operation = reset($operations);
    $messenger
      ->addMessage(t('An error occurred while processing @operation with arguments : @args', [
        '@operation' => $error_operation[0],
        '@args' => print_r($error_operation[0], TRUE),
      ]));
  }

  // Redirect the user to the system status report.
  return new RedirectResponse(Url::fromRoute('system.status')->setAbsolute()->toString());

}

/**
 * Function to update user_history records when a tracked field is added or
 * removed from the history processing.  This function is executed in a batch
 * context.
 *
 * @param integer $batch_size
 * @param array $add_fields
 * @param array $remove_fields
 * @param array $context
 *
 * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
 * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
 * @throws \Drupal\Core\Entity\EntityStorageException
 */
function user_history_update_tracked_history($batch_size, $add_fields, $remove_fields, &$context) {

  $entity_type = 'user_history';
  $values = [];
  $current_user = \Drupal::currentUser();

  $config = \Drupal::config('user_history.settings');
  // Get a list of fields that should be tracked by the user_history entity.
  $attached_fields = $config->get('attached_field');
  $tracked_fields = [];
  foreach ($attached_fields as $field_name => $tracked) {
    if ($tracked) {
      $tracked_fields[] = $field_name;
    }
  }

  // Initialise the sandbox values on the first call.
  if (empty($context['sandbox'])) {
    $context['sandbox'] = [];
    $context['sandbox']['total'] = \Drupal::entityQuery('user')->condition('uid', 0, '>')->count()->execute();
    $context['sandbox']['last'] = 0;
    $context['sandbox']['count'] = 0;
    $context['sandbox']['updated'] = 0;
    // Get current time for entitiy labels
    $now = new \Drupal\Core\Datetime\DrupalDateTime('now', drupal_get_user_timezone());
    $context['sandbox']['date'] = $now->format('Ymd-Hi');
  }

  // Retrieve the next set of user accounts
  $uids = \Drupal::entityQuery('user')
    ->condition('uid', $context['sandbox']['last'], '>')
    ->range(0, $batch_size)
    ->execute();

  $accounts = \Drupal::entityTypeManager()->getStorage('user')->loadMultiple($uids);

  /** @var \Drupal\user\Entity\User $account */
  foreach ($accounts as $account) {
    // Ignore the anonymous account.
    if (!$account->isAnonymous()) {

      // Set flag to initially suppress writing a user_history record to database.
      $user_history_update = FALSE;

      // Create a new user_history entity.
      $user_history = new UserHistory($values, $entity_type);

      // Empty arrays for list of new fields added or old fields removed from
      // the user_history record for this account.
      $new_fields = [];
      $old_fields = [];

      // Get a list of fields attached to the user_history entity.
      $field_definitions = \Drupal::service('entity_field.manager')->getFieldDefinitions('user_history', 'user_history');
      // Process each of these defined fields.
      foreach ($field_definitions as $field_name => $field_definition) {
        // Ignore base field definitions and process only attached fields
        /** \Drupal\field\Entity\FieldConfig $field_definition */
        if ($field_definition instanceof FieldConfig) {
          /** @var \Drupal\Core\Field\FieldItemList $field_item_list */
          $field_item_list = $account->get($field_name);

          // Record the value (if any) of the attached field in the user_history record?
          if (!empty($field_item_list)) {
            $field_type = $field_definition->getType();

            // Check if account has a value for the field being removed from tracking.
            if (in_array($field_name, $remove_fields) && $field_item_list->count() > 0) {
              // Yes, so set a flag to save a user_history record and record this as a removed field.
              $user_history_update = TRUE;
              $old_fields[] = $field_definitions[$field_name]->getLabel();
            }
            elseif (in_array($field_name, $add_fields) && $field_item_list->count() > 0) {
              // Account has a value for a field which is being added to tracking,
              // so extract the values and add to a user_history record.
              $history_values = user_history_get_tracked_field_value($field_item_list);
              if (!empty($history_values)) {
                // Set a flag to create a new user_history record this as a new attached field
                $user_history_update = TRUE;
                $user_history->set($field_name, $history_values);
                $new_fields[] = $field_definitions[$field_name]->getLabel();
              }
            }
            elseif (in_array($field_name, $tracked_fields) && $field_item_list->count() > 0) {
              // Account has a value for a field which was already being tracked,
              // so extract the values and add to a user_history record.
              $history_values = user_history_get_tracked_field_value($field_item_list);
              if (!empty($history_values)) {
                // Do not set a flag to create a new user_history record,
                // or record the field name as being added or removed.
                $user_history->set($field_name, $history_values);
              }
            }
          }
        }
      }

      $difference = [];
      if (!empty($new_fields)) {
        $difference[] = 'Add field values: ' . implode(', ', $new_fields);
      }
      if (!empty($old_fields)) {
        $difference[] = 'Remove field values: ' . implode(', ', $old_fields);
      }

      // If the user_history entry contains values for any attached fields,
      // then populate the base fields and save.
      if ($user_history_update) {
        // Set the label for this user_history entity
        $user_history->setLabel($account->id() . ': Modify - ' . $context['sandbox']['date']);

        // Set the modified time to the last changed time of the account
        $user_history->setCreatedTime($account->getChangedTime());
        // Set the modified_by uid value to current user.
        $user_history->setModifiedBy($current_user);
        // Record that this record was generated during tracked field update.
        $user_history->setAction('Modify tracked fields');
        // Record that this user entity exists in the database.
        $user_history->setUserDeleted(FALSE);

        $user_history->setUserId($account->id());
        $user_history->setUsername($account->getAccountname());
        $user_history->setUserPass($account->getPassword());
        $user_history->setUserMail($account->getEmail());
        $user_history->setUserTimezone($account->getTimeZone());
        $user_history->setUserStatus($account->isActive());
        $user_history->setUserRoles(implode('; ', $account->getRoles(TRUE)));
        $user_history->setUserCreated($account->getCreatedTime());
        $user_history->setUserChanged($account->getChangedTime());
        $user_history->setUserAccess($account->getLastAccessedTime());
        $user_history->setUserLogin($account->getLastLoginTime());
        $user_history->setUserInit($account->getInitialEmail());
        $user_history->setUserLangcode($account->language());
        $user_history->setUserPreferredLangcode($account->getPreferredLangcode());
        $user_history->setUserPreferredAdminLangcode($account->getPreferredAdminLangcode());

        // Record the reason for adding this user_history record.
        $user_history->setDifference(implode('. ', $difference));

        // Mark this user_history as a new entity.
        $user_history->enforceIsNew(TRUE);
        $user_history->save();
        $context['sandbox']['updated']++;
      }

    }

    $context['sandbox']['last'] = $account->id();
    $context['sandbox']['count']++;
  }

  if ($context['sandbox']['count'] < $context['sandbox']['total']) {
    $context['finished'] = $context['sandbox']['count'] / $context['sandbox']['total'];
    if ($context['finished'] > 0.99) {
      $context['finished'] = 0.99;
    }
    $context['message'] = t('Processed %count of %total user accounts.', [
      '%count' => $context['sandbox']['count'],
      '%total' => $context['sandbox']['total'],
    ]);
  }
  else {
    $context['finished'] = 1;
    $context['message'] = t('Finished processing %total user accounts. Added %num new user history records', [
      '%total' => $context['sandbox']['total'],
      '%num' => $context['sandbox']['updated'],
    ]);
    $context['results']['total'] = $context['sandbox']['total'];
    $context['results']['updated'] = $context['sandbox']['updated'];
  }

}

function user_history_finished_update_attached_fields($success, $results, $operations) {

  $messenger = \Drupal::messenger();

  if ($success) {
    // Flag the initialisation as complete.
    \Drupal::state()->set('user_history.attached_field_update_required',FALSE);
    // Display a completion message to the user.
    \Drupal::messenger()->addMessage(t('Added %num new user_history records', ['%num' => $results['updated']]));
  }
  else {
    // An error occurred.
    // $operations contains the operations that remained unprocessed.
    $error_operation = reset($operations);
    $messenger
      ->addMessage(t('An error occurred while processing @operation with arguments : @args', [
        '@operation' => $error_operation[0],
        '@args' => print_r($error_operation[0], TRUE),
      ]));
  }

  // Redirect to the system status report page.
  return new RedirectResponse(Url::fromRoute('system.status')->setAbsolute()->toString());

}

/**
 * Function to archive user_history records. This function is executed in a batch
 * context.
 *
 * @param array $archive_date
 * @param integer $min_records
 * @param integer $max_records
 * @param string $file_format
 * @param integer $batch_size
 * @param array $context
 *
 * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
 * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
 * @throws \Drupal\Core\Entity\EntityStorageException
 */
function user_history_archive_records($archive_date, $min_records, $max_records, $file_format, $batch_size, &$context) {

  // Convert the $archive_date array to a timestamp
  if (!empty($archive_date)) {
    $archive_date = DrupalDateTime::createFromFormat('Y-m-d', $archive_date);
    $archive_timestamp = $archive_date->format('U');
  }
  else {
    $archive_timestamp = 0;
  }

  // Initialise the sandbox values on the first call.
  // The batch process will iterate through all user ids.
  if (empty($context['sandbox'])) {
    $context['sandbox'] = [];
    $context['sandbox']['total'] = \Drupal::entityQuery('user')
      ->condition('uid', 0, '>')
      ->count()
      ->execute();

    $config = \Drupal::config('user_history.settings');

    $fileSystem = \Drupal::service('file_system');
    // Check if the private file stream wrapper is available for use.
    $filepath = $fileSystem->validScheme('private') ? 'private://' : 'public://';
    // Append the specified archive file path to the file scheme.
    $filepath .= $config->get('archive.directory') ?? 'user_history/archive/';
    // Get the file name and append the current date/time.
    $filename = $config->get('archive.filename') ?? 'user-history-archive';
    $filename .= '-' . date('Ymd') . '-' . date('Hi') . '.' . $file_format;

    // Ensure the filepath exists
    if (!\Drupal::service('file_system')->prepareDirectory($filepath, FileSystemInterface::CREATE_DIRECTORY)) {
      $context['message'] = t('ERROR - Unable to access or create the specified archive directory');
      $context['results']['total'] = 0;
      $context['results']['archived'] = 0;
      $context['finished'] = 1;
      return;
    }

    $context['sandbox']['last'] = 0;
    $context['sandbox']['count'] = 0;
    $context['sandbox']['archived'] = 0;
    $context['sandbox']['ids'] = [];
    $context['sandbox']['filepath'] = $filepath;
    $context['sandbox']['filename'] = $filename;
    $header_write = FALSE;
    $context['sandbox']['header_write'] = $header_write;

    // Initialise the output file.
    $archive_file = fopen($filepath . $filename, "w");
  }
  else {
    // Recover the values of variables from the context sandbox.
    $filepath = $context['sandbox']['filepath'];
    $filename = $context['sandbox']['filename'];
    $header_write = $context['sandbox']['header_write'];

    // Re-open the output file for subsequent batches.
    $archive_file = fopen($filepath . $filename, "a");
  }

  // Retrieve the next set of user ids
  $uids = \Drupal::entityQuery('user')
    ->condition('uid', $context['sandbox']['last'], '>')
    ->range(0, $batch_size)
    ->execute();

  foreach ($uids as $uid) {

    // Count the user_history records for this account.
    $num_records = \Drupal::entityQuery('user_history')
      ->condition('user_id', $uid)
      ->count()
      ->execute();

    // Ensure that there are at least $min_records history records for this
    // account before considering any archiving.
    if ($num_records > $min_records) {
      $uhids = [];
      // Find records that are earlier then the archive date
      $ids = \Drupal::entityQuery('user_history')
        ->condition('user_id', $uid)
        ->condition('created', $archive_timestamp, '<')
        ->sort('created', 'ASC')
        ->execute();
      if (count($ids) > 0) {
        // Check that we are not archiving / deleting too many records.
        $remaining = $num_records - count($ids);
        while ($remaining < $min_records) {
          $keep = array_pop($ids);
          $remaining++;
        }
        $uhids = $ids;
      }
      // Find old records that are in excess of the max retained value
      if ($num_records > $max_records) {
        $ids = \Drupal::entityQuery('user_history')
          ->condition('user_id', $uid)
          ->sort('created', 'ASC')
          ->range(0, $num_records - $max_records)
          ->execute();
        $uhids = array_unique($uhids + $ids);
      }

      // Now load and archive the selected records
      $user_histories = \Drupal::entityTypeManager()->getStorage('user_history')->loadMultiple($uhids);

      foreach ($user_histories as $user_history) {

        $uh_field_info = user_history_get_user_history_fields();

        switch ($file_format) {
          case 'txt':
            $record = user_history_values_to_txt($user_history) . PHP_EOL;
            break;

          case 'csv':
            if (!$header_write) {
              $max_cardinality = $config->get('archive.max_cardinality');
              $headers = user_history_csv_headers($uh_field_info, $max_cardinality);
              fwrite($archive_file, implode(',', $headers) . PHP_EOL);
              $header_write = TRUE;
            }
            $record = user_history_values_to_csv($user_history, $uh_field_info, $max_cardinality) . PHP_EOL;
            break;

          case 'xml':
            if (!$header_write) {
              $headers = '<?xml version="1.0"?><user_histories>';
              fwrite($archive_file, $headers);
              $header_write = TRUE;
            }
            $record = user_history_values_to_xml($user_history);
            // Strip the xml tag from the front of the record
            $record = preg_replace('/^<\?xml(.*)\?>/', '', $record);
            break;

          case 'json':
            $record = user_history_values_to_json($user_history);
            break;

          default:
            $record = NULL;
        }
        fwrite($archive_file, $record);
        // Update count of archived user history records
        $context['sandbox']['archived']++;
        // Save the id of the archived user_history record.
        $context['sandbox']['ids'][] = $user_history->id();
      }

      // Have processed all the user history records for this account.
    }
    // Update count of uids processed.
    $context['sandbox']['count']++;
    $context['sandbox']['last'] = $uid;
  }

  // Finished this batch of uids, so close archive file.
  fclose($archive_file);
  // Save header_write state
  $context['sandbox']['header_write'] = $header_write;

  // Check if all uids have been processed or whether there are more batches to go.
  if ($context['sandbox']['count'] < $context['sandbox']['total']) {
    $context['finished'] = $context['sandbox']['count'] / $context['sandbox']['total'];
    if ($context['finished'] > 0.99) {
      $context['finished'] = 0.99;
    }
    $context['message'] = t('Processed %count of %total user accounts.', [
      '%count' => $context['sandbox']['count'],
      '%total' => $context['sandbox']['total'],
    ]);
  }
  else {
    $context['finished'] = 1;

    // Add closing tag to XML file if necessary
    if ($file_format == 'xml') {
      $archive_file = fopen($filepath . $filename, "a");
      fwrite($archive_file, '</user_histories>');
      fclose($archive_file);
    }

    $context['message'] = t('Finished processing %total user accounts. Archived %num user history records', [
      '%total' => $context['sandbox']['total'],
      '%num' => $context['sandbox']['archived'],
    ]);
    $context['results']['total'] = $context['sandbox']['total'];
    $context['results']['archived'] = $context['sandbox']['archived'];
    $context['results']['ids'] = $context['sandbox']['ids'];
  }
}

/**
 * Function to archive user_history records. This function is executed in a batch
 * context.
 *
 * @param array $archive_date
 * @param integer $min_records
 * @param integer $max_records
 * @param integer $batch_size
 * @param array $context
 *
 * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
 * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
 * @throws \Drupal\Core\Entity\EntityStorageException
 */
function user_history_delete_records($archive_date, $min_records, $max_records, $batch_size, &$context) {

  // Initialise the sandbox values on the first call.
  // The batch process will process the set of archived user_history ids.
  if (empty($context['sandbox'])) {
    $context['sandbox'] = [];
    $context['sandbox']['total'] = count($context['results']['ids']);
    $context['sandbox']['count'] = 0;
    $context['sandbox']['deleted'] = 0;
  }

  // Extract the next set of user_history ids
  $uhids = array_slice($context['results']['ids'], $context['sandbox']['count'], $batch_size);

  // Now load and archive the selected records
  $user_histories = \Drupal::entityTypeManager()->getStorage('user_history')->loadMultiple($uhids);

  foreach ($user_histories as $user_history) {
    $user_history->delete();
    // Update count of archived user history records
    $context['sandbox']['deleted']++;
  }

  // Update count of user_history ids processed.
  $context['sandbox']['count'] = $context['sandbox']['count'] + count($uhids);

  // Check if all ids have been processed or whether there are more batches to go.
  if ($context['sandbox']['count'] < $context['sandbox']['total']) {
    $context['finished'] = $context['sandbox']['count'] / $context['sandbox']['total'];
    if ($context['finished'] > 0.99) {
      $context['finished'] = 0.99;
    }
    $context['message'] = t('Processed %count of %total user_history records.', [
      '%count' => $context['sandbox']['count'],
      '%total' => $context['sandbox']['total'],
    ]);
  }
  else {
    $context['finished'] = 1;
    $context['message'] = t('Finished processing %total user_history records. Deleted %num user history records', [
      '%total' => $context['sandbox']['total'],
      '%num' => $context['sandbox']['deleted'],
    ]);
    $context['results']['total2'] = $context['sandbox']['total'];
    $context['results']['deleted'] = $context['sandbox']['deleted'];
  }
}

/**
 * Function to handle completion of batch process to archive and optionally
 * delete user history records.  Display a message to user and write to log.
 *
 * @param array $success
 * @param array $results
 * @param array $operations
 *
 * @return \Symfony\Component\HttpFoundation\RedirectResponse
 */
function user_history_finished_archive_records($success, $results, $operations) {

  if ($success) {
    // Display a completion message to the user.
    $message = t('Processed %total user accounts and archived %num user history records',
      ['%total' => $results['total'], '%num' => $results['archived']]);

    \Drupal::messenger()->addMessage($message);
    \Drupal::logger('user_history')->notice($message);
  }
  else {
    // An error occurred.
    // $operations contains the operations that remained unprocessed.
    $error_operation = reset($operations);
    $message = t('An error occurred while processing @operation with arguments : @args', [
        '@operation' => $error_operation[0],
        '@args' => print_r($error_operation[0], TRUE),
      ]);

    \Drupal::messenger()->addMessage($message);
    \Drupal::logger('user_history')->error($message);
  }

  // Redirect to the system status report page.
  return new RedirectResponse(Url::fromRoute('system.status')->setAbsolute()->toString());
}


/**
 * Function to return an array of values that can be used as a header record
 * for a CSV archive file.
 *
 * @param array $uh_field_info
 * @param integer $max_cardinality
 *
 * @return array
 */
function user_history_csv_headers($uh_field_info, $max_cardinality) {

  $headers = [];

  foreach ($uh_field_info as $field_name => $field_info) {
    $cardinality = $field_info['cardinality'];
    unset($field_info['cardinality']);
    if ($cardinality == -1 || $cardinality > $max_cardinality) {
      $headers[] = $field_name . ':serialized';
    }
    elseif ($cardinality == 1) {
      foreach ($field_info as $property) {
        $headers[] = $field_name . ':' . $property;
      }
    }
    else {
      for ($count = 0; $count < $cardinality; $count++) {
        foreach ($field_info as $property) {
          $headers[] = $field_name . ':' . $count . ':' . $property;
        }
      }
    }
  }
  return $headers;
}

/**
 * Function to convert a user_history entity to a string for adding to a TXT
 * archive file.  The user_history entity is first converted to an array and
 * then the array is serialized and massaged to deal with embedded newline
 * characters.
 *
 * @param \Drupal\user_history\Entity\UserHistory $user_history
 *
 * @return string
 */
function user_history_values_to_txt($user_history) {
  // Convert user_history entity to an array of values
  $uh_values = $user_history->toArray();
  // Generate a serialised array, and encode CR/LF chars.
  return str_replace(["\n", "\r"], ['\n', '\r'], serialize($uh_values));
}

/**
 * Function to convert a user_histroy entity into a string for adding to a CSV
 * archive file. The user_history entity is first converted to an array and
 * then the elements of the array are processed to generate CSV columns matching
 * the csv header that has been generated.
 *
 * @param \Drupal\user_history\Entity\UserHistory $user_history
 * @param array $uh_field_info
 * @param integer $max_cardinality
 *
 * @return string
 */
function user_history_values_to_csv($user_history, $uh_field_info, $max_cardinality) {
  // Convert user_history entity to an array of values
  $uh_values = $user_history->toArray();

  $return = [];
  foreach ($uh_field_info as $field_name => $field_info) {
    $cardinality = $field_info['cardinality'];
    unset($field_info['cardinality']);
    $field_values = $uh_values[$field_name];
    if ($cardinality == -1 || $cardinality > $max_cardinality) {
      $return[] = serialize($field_values);
    }
    else {
      for ($count = 0; $count < $cardinality; $count++) {
        if (isset($field_values[$count])) {
          foreach ($field_info as $property) {
            $return[] = $field_values[$count][$property];
          }
        }
        else {
          foreach ($field_info as $property) {
            $return[] = NULL;
          }
        }
        $count++;
      }
    }
  }

  // Add quotes to any string values and escape commas and quote marks.
  foreach ($return as $key => $value) {
    if (!is_numeric($value) && !is_null($value)) {
      $return[$key] = '"' . str_replace([',', '"'], ['\,', '\"'], trim($value, '"')) . '"';
    }
  }

  // Encode any CR or LF chars present in the return value.
  return str_replace(["\n", "\r"], ['\n', '\r'], implode(',', $return));

}

/**
 * Function to convert a user_history entity into a serialized XML data structure
 * using the core normalizations and serialization functions.
 *
 * @param \Drupal\user_history\Entity\UserHistory $user_history
 *
 * @return mixed
 */
function user_history_values_to_xml($user_history) {
  return \Drupal::service('serializer')->serialize($user_history, 'xml');
}

/**
 * Function to convert a user_history entity into a serialized JSON data structure
 * using the core normalizations and serialization functions.
 *
 * @param \Drupal\user_history\Entity\UserHistory $user_history
 *
 * @return mixed
 */
function user_history_values_to_json($user_history) {
  return \Drupal::service('serializer')->serialize($user_history, 'json');
}

/**
 * Function to get the user_history fields and return an array spcifying the
 * properties and cardinality of each field.
 *
 * @return array
 */
function user_history_get_user_history_fields() {
  $return = [];
  $entity_field_manager = \Drupal::service('entity_field.manager');
  $field_definitions = $entity_field_manager->getFieldDefinitions('user_history', 'user_history');
  foreach ($field_definitions as $field_name => $field_definition) {

    if ($field_definition instanceof BaseFieldDefinition) {
      $return[$field_name][] = $field_definition->getMainPropertyName();
      $return[$field_name]['cardinality'] = $field_definition->getCardinality();
    }
    else {
      // Field added through UI or programatically.
      /** @var \Drupal\field\Entity\FieldConfig $field_config */
      $field_config = $field_definition;
      $field_storage_defn = $field_config->getFieldStorageDefinition();
      foreach (array_keys($field_storage_defn->getColumns()) as $property) {
        $return[$field_name][] = $property;
      }
      $return[$field_name]['cardinality'] = $field_storage_defn->getCardinality();
    }
  }
  return $return;
}
